#   Element: Software interpretable representation of an XML element.
#   Copyright (C) 1999  Daniel Arbuckle
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    
#   As a special exception, this code may be linked with code not covered
#   by the GNU General Public License, as long as the license covering the
#   other code does not restrict or modify the provisions governing the
#   use of this code.
#
#    Daniel Arbuckle can be contacted by email at djarb@wvc-omak.ctc.edu

import UserDict
import string
import JabberUtil
import re

class Element(UserDict.UserDict):
    """This class implements a sort of document object model. It
    represents a tag in an XML document, which can have children,
    attributes, data, and a namespace."""
    
    def __init__(self, type):
        UserDict.UserDict.__init__(self)
        self.type = type #The tag. e.g. for <baz foo="bar"/>, it's "baz"
        self.cdata = "" #Between tag data (character data), NOT real XML cdata
        self.children = [] #List of nested elements
    
    def __setitem__(self, key, value):
        """Allows the [] index syntax to be used on elements. When
        passed a string, this operation returns the attribute of that
        name.""" 
        UserDict.UserDict.__setitem__(self, JabberUtil.split_name(key)[1], value)
                             
    def update(self, other):
        """other must support the mapping interface. Adds all the
        items in other to self."""
        for (k, v) in other.items():
            self[JabberUtil.split_name(k)[1]] = v
            
    def count_children(self):
        """Returns the number of child elements this element has."""
        return len(self.children)
    
    def has_child(self, name):
        """returns tryue if any child of the element matches the
        parameter.""" 
        if type(name) == type(""):
            regex = re.compile(name)
        else:
            regex = name
            
        for c in self.children:
            if regex.match(c.type):
                return 1
        return 0
    
    def has_child_not(self, name):
        """Returns true if any child of the element does not match the
        parameter."""
        if type(names) == type(""):
            regex = re.compile(name)
        else:
            regex = name

        for c in self.children:
            if not regex.match(c.type):
                return 1
        return 0
    
    def get_children(self, children = None):
        """Returns a list of all the children matching the parameter,
        or all children if the parameter is omitted."""
        if children == None:
            return self.children[:] #Return a clone list, to protect my copy

        if type(children) == type(""):
            regex = re.compile(children)
        else:
            regex = children
            
        ret = []
        for c in self.children:
            if regex.match(c.type):
                ret.append(c)
        return ret

    def get_children_not(self, children):
        """Returns a list of all children that don't match the
        parameter.""" 
        if type(children) == type(""):
            regex = re.compile(children)
        else:
            regex = children
            
        ret = []
        for c in self.children:
            if not regex.match(c.type):
                ret.append(c)
        return ret
    
    def add_child(self, child):
        """Adds a new child element to self"""
        self.children.append(child)
        
    #Transforms the element back into XML data
    #BUG: attributes are forced to be in the same namespace as the tag. I
    #     don't expect this to be a problem in the context of Jabber XML
    #     streams.
    def output(self, depth = 0, namespace = ""):
        textlist = []
        if depth >= 0: #If depth is less than zero, pretty printing is off
            textlist.append("\t" * depth) #Indent to the right level
        name = JabberUtil.split_name(self.type)
        textlist.append("<%s" % name[1]) #Begin the tag
        if name[0] and name[0] != namespace:
            textlist.append(' xmlns="%s"' % name[0])
        for i in self.items(): #Loop through attributes
            if i[1] == None: continue #Skip empties
            attr = JabberUtil.split_name(i[0])
            textlist.append(' %s="%s"' % (attr[1], i[1])) #Print attribute key/value pair
        if len(self.children) == 0 and len(self.cdata) == 0: #Nothing nested
            textlist.append("/>") #End the tag
            if depth >= 0:
                textlist.append("\n")
            return string.join(textlist, '')
        textlist.append(">") #End the tag for the nested case
        if depth >= 0:
            textlist.append("\n")
        if len(self.cdata): #It has nested character data
            if depth >= 0:
                outcdata = string.strip(self.cdata) #Get rid of white space
                if len(outcdata): #If there's something left, print it
                    textlist.append("\t" * (depth + 1))
                    textlist.append(outcdata)
                    textlist.append("\n")
            else:
                textlist.append(self.cdata) #Not pretty, so print the cdata literally
        for e in self.children: #Loop through nested elements
            if depth >= 0:
                textlist.append(e.output(depth + 1, name[0]))
            else:
                textlist.append(e.output(depth, name[0]))
        if depth >= 0:
            textlist.append("\t" * depth)
        textlist.append("</%s>" % name[1]) #End the tag
        if(depth >= 0):
            textlist.append("\n")
        return string.join(textlist, '')
