#   Jabber.py: Implements the Jabber protocol, on top of JabberParser
#   Copyright (C) 1999  Daniel Arbuckle
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    
#   As a special exception, this code may be linked with code not covered
#   by the GNU General Public License, as long as the license covering the
#   other code does not restrict or modify the provisions governing the
#   use of this code.
#
#    Daniel Arbuckle can be contacted by email at djarb@wvc-omak.ctc.edu

import Element
import JabberParser

import socket
import string
import time
import sha

# There are only two requirements on packets: they must implement a
# constructor which accepts an Element.Element as it's only parameter, and
# they must implement an element_tree() method returning Element.Element
from JabberPackets import *

NoPackets = "There are no available packets."
BadSend = "Cannot send this type of packet"

# Some Python interpreters are compiled for multithreading, some
# aren't. In order to operate correctly in threaded environments,
# PyJabber needs to take advantage mutexen, synchronized queues,
# etc. Rather than put if statements all through the code, I simply
# define stub versions of the threading classes on pythons which don't
# have threading.
try: 
    from threading import *
    from Queue import *
    jabber_is_threaded = 1
except:
    jabber_is_threaded = 0
    class Thread:
        "Fake thread class for use in nonthreaded environments"
        def __init__(self, group=None, target=None,
                     name=None, args=(), kwargs={}):
            pass

        def start(self):
            pass

        def setDaemon(self, daemonic):
            pass

    class Lock:
        "Fake lock class for use in nonthreaded environments"
        def acquire(self, blocking = 1):
            pass

        def release(self):
            pass
        
    class Queue:
        "Fake queue class for use in nonthreaded environments"
        def __init__(self, maxsize = 0):
            self.__data = []

        def qsize(self):
            return len(self.__data)

        def empty(self):
            return len(self.__data) == 0

        def put(self, item, block = 1):
            self.__data.append(item)

        def get(self, block = 1):
            if len(self.__data):
                raise NoPackets
            ret = self.__data[0]
            self.__data = self.__data[1:]
            return ret
# End of the code to deal with threads

class UnrecognisedPacket:
    "Represents a recieved packet for which no handler is registered"
    def __init__(self, arg):
        if isinstance(arg, Element):
            self.msg = "Unrecognised packet type: " + arg.type
            self.tree = arg
        else:
            self.msg = arg
            self.tree = None

    def element_tree(self):
        raise BadSend, "UnrecognisedPacket"
    
class PacketQueue:
    """The low level API to the Jabber stream protocol. The primary
    interface to PacketQueue objects is via their get() and feed()
    methods. feed() gives the parser more data to process, while get
    returns the next available packet."""
    
    def __init__(self):
        self.__callback = None
        self.__parser = JabberParser.JabberParser(1)
        self.__packets = {}
        self.__avail = Queue(0)        
        self.__parser.use_default_handler(self.__incoming_packet_handler)
        
        auto_register_jabber_packets(self)        
        
    def __incoming_packet_handler(self, tree):
        if not self.__packets.has_key(tree.type):
            self.__avail.put(UnrecognisedPacket(tree))
        else: self.__avail.put(self.__packets[tree.type](tree))
        if callable(self.__callback): self.__callback()
        
    def feed(self, data):
        """This function should be called any time there are new data
        which should be processed by PyJabber. For example, just after
        a call to recv()"""
        self.__parser.feed(data)
        
    def get_header(self, to):
        """Returns the string to use as the opening tag of your jabber
        session."""
        return '<stream:stream to="%s" xmlns="jabber:client" xmlns:stream="http://etherx.jabber.org/streams">' % to
    
    def get(self):
        """Called to get the next packet from the incoming queue. If
        you are operating in a threaded environment, this will block
        until a packet is available. In a nonthreaded environment, a
        Jabber.NoPackets exception is thrown in the event of an empty
        queue."""
        return self.__avail.get() #Queue handles its own synchronization

    def available(self):
        "Returns the number of Packets in the incoming queue."
        return self.__avail.qsize()

    def register_packet_type(self, name, packet):
        """Called to add a new packet type to the list of understood
        packets. First parameter is a string of the format
        'namespace tag' which identifies the namespace and tag which
        the packet type handles. Second parameter will be called to
        construct a new packet whenever that tag is found."""
        self.__packets[name] = packet

    def set_callback(self, cb):
        """Allows you to set a callback which will be triggered on the 
        reception of a new packet. In a multithreaded environment,
        there are better ways of dealing with the problem of
        notification."""
        self.__callback = cb

def generate_digest(id, password):
    """When passed a session id and a password, returns a digest
    suitable for use with Jabber digest authentication."""
    base = "%s%s" % (id, password)
    dig = sha.new(base).hexdigest()
    return dig

## The things below here are not strictly necessary for implementing 
## a Jabber program. They are simply utilities, inteded to simplify
## a common case, and to serve as example code.

class SimpleClient(PacketQueue):
    """This class implements some of the work of creating a simple
    Jabber client on a socket based network"""
    
    def __init__(self, server, port = 5222, use_threads = 1,
                 in_debug = None, out_debug = None):
        if type(server) == type("string"):
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            addr = string.split(server, ":")
            try:
                self.__sock.connect((addr[0], string.atoi(addr[1])))
            except IndexError:
                self.__sock.connect((addr[0], port))

            if use_threads:
                self.__sock.setblocking(0)
        else:
            self.__sock = server
            
        PacketQueue.__init__(self)
        
        self.in_debug = in_debug
        self.out_debug = out_debug
        
        self.__sock_lock = Lock()

        self.__sock_lock.acquire()
        self.__sock.send(self.get_header(addr[0]))
        self.__sock_lock.release()
        
        if use_threads:
            self.thread = Thread(target = self.__update_periodically)
            self.thread.setDaemon(1)
            self.thread.start()

    def put(self, packet):
        """Call to send a packet to the server."""
        self.__sock_lock.acquire()
        e = packet.element_tree()
        s = e.output(-1)
        if self.out_debug: self.out_debug.send(s)
        self.__sock.send(s)
        self.__sock_lock.release()
        
    def read_avail(self):
        """If data are available, reads them and stuffs them into the
        parser. Otherwise simply returns. This function never
        blocks.""" 
        self.__sock_lock.acquire()
        try:
            s = self.__sock.recv(1000)
        except socket.error:
            s = None
        self.__sock_lock.release()

        if s:
            if self.in_debug: self.in_debug.send(s)
            self.feed(s)
        
    def __update_periodically(self):
        """If threading is available and use_threads is true, this
        function will execute in a seperate thread and handle the
        actual reading of data from the socket. Otherwise, it will not
        be called, and you will have to call read_avail manually from
        time to time."""
        while 1:
            self.read_avail()
            time.sleep(0.1)
