#   JabberPackets.py: Protocol and version specific Jabber packet wrappers
#   Copyright (C) 1999  Daniel Arbuckle
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    
#   As a special exception, this code may be linked with code not covered
#   by the GNU General Public License, as long as the license covering the
#   other code does not restrict or modify the provisions governing the
#   use of this code.
#
#    Daniel Arbuckle can be contacted by email at djarb@wvc-omak.ctc.edu

import re
import FileSocket
import JabberUtil
from Element import *

try:
    from cStringIO import *
except:
    from StringIO import *

XMLFailure = "Conversion to XML failed"
ParseFailure = "Conversion from XML failed"
BadUsage = "Bad usage"

class DefaultExtension:
    def __init__(self, tree):
        self.tree = tree

    def element_tree(self):
        raise XMLFailure, "Cannot convert a DefaultExtension object"
    
class Ident:
    NS = "x:ident"
    TO_TAG = NS + " to"
    FROM_TAG = NS + " from"
    CC_TAG = NS + " cc"
    FORWARDEDBY_TAG = NS + " forwardedby"
    REPLYTO_TAG = NS + " replyto"
    TO = re.compile(TO_TAG)
    FROM = re.compile(FROM_TAG)
    CC = re.compile(CC_TAG)
    FORWARDEDBY = re.compile(FORWARDEDBY_TAG)
    REPLYTO = re.compile(REPLYTO_TAG)
    
    def __init__(self, tree = None):
        if tree == None:
            self.__to = []
            self.__from = []
            self.__cc = []
            self.__forwardedby = []
            self.__replyto = []
        else:
            for e in tree.get_children(Ident.TO):
                self.__to.append(e.cdata)

            for e in tree.get_children(Ident.FROM):
                self.__from.append(e.cdata)

            for e in tree.get_children(Ident.CC):
                self.__cc.append(e.cdata)

            for e in tree.get_children(Ident.FORWARDEDBY):
                self.__forwardedby.append(e.cdata)

            for e in tree.get_children(Ident.REPLYTO):
                self.__replyto.append(e.cdata)

    def element_tree(self):
        tree = Element(Ident.NS + " x")

        for e in self.__to:
            s = Element(Ident.TO_TAG)
            s.cdata = e
            tree.add_child(s)
            
        for e in self.__from:
            s = Element(Ident.FROM_TAG)
            s.cdata = e
            tree.add_child(s)

        for e in self.__cc:
            s = Element(Ident.CC_TAG)
            s.cdata = e
            tree.add_child(s)

        for e in self.__forwardedby:
            s = Element(Ident.FORWARDEDBY_TAG)
            s.cdata = e
            tree.add_child(s)

        for e in self.__replyto:
            s = Element(Ident.REPLYTO_TAG)
            s.cdata = e
            tree.add_child(s)

        return tree

    def add_to(self, to_):
        self.__to.append(to_)

    def get_to(self):
        return self.__to

    def add_from(self, from_):
        self.__from.append(from_)

    def get_from(self):
        return self.__from

    def add_cc(self, cc_):
        self.__cc.append(cc_)

    def get_cc(self):
        return self.__cc

    def add_forwardedby(self, f_):
        self.__forwardedby.append(f_)

    def get_forwardedby(self):
        return self.__forwardedby

    def add_replyto(self, replyto_):
        self.__replyto.append(replyto_)

    def get_replyto(self):
        return self.__replyto

class Delay:
    NS = "x:delay"
    
    def __init__(self, tree = None):
        if tree == None:
            self.__from = ""
            self.__stamp = ""
            self.__body = ""
        else:
            if tree.has_key("from"): self.__from = tree['from']
            else: self.__from = ""

            if tree.has_key("stamp"): self.__stamp = tree['stamp']
            else: self.__stamp = ""

            self.__body = tree.cdata

    def element_tree(self):
        tree = Element(Ident.NS + " x")
        if self.__from: tree['from'] = self.__from
        if self.__stamp: tree['stamp'] = self.__stamp
        tree.cdata = self.__body
        return tree

    def set_from(self, from_):
        self.__from = from_

    def get_from(self):
        return self.__from

    def set_stamp(self, stamp_):
        self.__stamp = stamp_

    def get_stamp(self):
        return self.__stamp

    def set_text(self, text_):
        self.__body = text_

    def get_text(self):
        return self.__body
    
class Stream:
    TAG = "http://etherx.jabber.org/streams stream"

    def __init__(self, tree = None):
        if not tree:
            self.__id = ""
            self.__from = ""
        else:
            if tree.has_key("id"): self.__id = tree["id"]
            else: raise ParseFailure, "stream without id"

            if tree.has_key("from"): self.__from = tree["from"]
            else: raise ParseFailure, "stream without from"

    def get_id(self):
        return self.__id

    def get_from(self):
        return self.__from
    
    def element_tree(self):
        raise BadUsage, "cannot send a stream tag over an open stream."
            
class Presence:
    TAG = "jabber:client presence"
    STATUS_TAG = "jabber:client status"
    PRIORITY_TAG = "jabber:client priority"
    META_TAG = "jabber:client meta"
    SHOW_TAG = "jabber:client show"
    LOC_TAG = "jabber:client loc"

    STATUS = re.compile(STATUS_TAG)
    PRIORITY = re.compile(PRIORITY_TAG)
    META = re.compile(META_TAG)
    SHOW = re.compile(SHOW_TAG)
    LOC = re.compile(LOC_TAG)
    X = re.compile(r"\S+ x")
    
    extension_packets = {}
    
    def __init__(self, tree = None):
        if not tree:
            self.__to = ""
            self.__from = ""
            self.__type = ""
            self.__status = ""
            self.__priority = ""
            self.__meta = ""
            self.__show = ""
            self.__loc = ""
            self.__x = []
        else:
            if tree.has_key("to"): self.__to = tree["to"]
            else: self.__to = ""
            if tree.has_key("from"): self.__from = tree["from"]
            else: self.__from = ""
            if tree.has_key("type"): self.__type = tree["type"]
            else: self.__type = ""

            if tree.has_child(Presence.STATUS):
                self.__status = tree.get_children(Presence.STATUS)[0].cdata
            else: self.__status = ""
            
            if tree.has_child(Presence.PRIORITY):
                self.__priority = tree.get_children(Presence.PRIORITY)[0].cdata
            else: self.__priority = ""
            
            if tree.has_child(Presence.META):
                self.__meta = tree.get_children(Presence.META)[0].cdata
            else: self.__meta = ""
            
            if tree.has_child(Presence.SHOW):
                self.__show = tree.get_children(Presence.SHOW)[0].cdata
            else: self.__show = ""
            
            if tree.has_child(Presence.LOC):
                self.__loc = tree.get_children(Presence.LOC)[0].cdata
            else: self.__loc = ""
            
            self.__x = []
            xen = tree.get_children(Presence.X)
            for x in xen:
                if Presence.extension_packets.has_key(x.type):
                    self.__x.append(Presence.extension_packets[x.type](x))
                else:
                    self.__x.append(DefaultExtension(x))

    def element_tree(self):
        tree = Element(Presence.TAG)

        if self.__to: tree['to'] = self.__to

        if self.__from: tree['from'] = self.__from

        if self.__type: tree['type'] = self.__type

        if self.__status:
            e = Element(Presence.STATUS_TAG)
            e.cdata = self.__status
            tree.add_child(e)

        if self.__priority:
            e = Element(Presence.PRIORITY_TAG)
            e.cdata = self.__priority
            tree.add_child(e)

        if self.__meta:
            e = Element(Presence.META_TAG)
            e.cdata = self.__meta
            tree.add_child(e)
        
        if self.__show:
            e = Element(Presence.SHOW_TAG)
            e.cdata = self.__show
            tree.add_child(e)
            
        if self.__loc:
            e = Element(Presence.LOC_TAG)
            e.cdata = self.__loc
            tree.add_child(e)

        for x in self.__x:
            tree.add_child(x.element_tree())

        return tree

            
    def set_to(self, to_):
        self.__to = to_
    
    def get_to(self):
        return self.__to
    
    def set_from(self, from_):
        self.__from = from_
    
    def get_from(self):
        return self.__from
    
    def set_type(self, type_):
        self.__type = type_
    
    def get_type(self):
        return self.__type
    
    def set_status(self, status_):
        self.__status = status_
    
    def get_status(self):
        return self.__status
    
    def set_priority(self, pri_):
        if type(pri_) != type(""): self.__priority = `pri_`
        else: self.__priority = pri_
    
    def get_priority(self):
        return int(self.__priority)
    
    def set_meta(self, meta_):
        self.__meta = meta_
    
    def get_meta(self):
        return self.__meta
    
    def set_show(self, show_):
        self.__show = show_
    
    def get_show(self):
        return self.__show
    
    def set_loc(self, loc_):
        self.__loc = loc_
    
    def get_loc(self):
        return self.__loc
    
    def add_extension(self, x_):
        self.__x.append(x_)
    
    def get_extensions(self):
        return self.__x
    
class IQ:
    TAG = "jabber:client iq"

    ERROR_TAG = "jabber:client error"
    ERROR = re.compile(ERROR_TAG)
    
    def __init__(self, tree = None):
        if not tree:
            self.__to = ""
            self.__from = ""
            self.__id = ""
            self.__type = ""
            self.__error = None
            self.__query = None
        else:
            if tree.has_key("to"): self.__to = tree["to"]
            else: self.__to = ""

            if tree.has_key("from"): self.__from = tree["from"]
            else: self.__from = ""

            if tree.has_key("id"): self.__id = tree["id"]
            else: self.__id = ""

            if tree.has_key("type"): self.__type = tree["type"]
            else: raise ParseFailure, "IQ without type"

            e = tree.get_children(IQ.ERROR)
            if len(e):
                self.__error = (e[0]["code"], e[0].cdata)
            else:
                self.__error = None

            q = tree.get_children()

            if len(q) == 0:
                self.__query = None
            else:
                self.__query = q[0]
            

    def element_tree(self):
        tree = Element(IQ.TAG)
        
        if self.__to: tree['to'] = self.__to

        if self.__from: tree['from'] = self.__from

        if self.__id: tree['id'] = self.__id

        if self.__type: tree['type'] = self.__type
        else: tree['type'] = "get"

        if self.__error:
            e = Element(IQ.ERROR_TAG)
            e['code'] = self.__error[0]
            c.cdata = self.__error[1]
            tree.add_child(e)

        tree.add_child(self.__query)

        return tree
    
    def set_to(self, to_):
        self.__to = to_

    def get_to(self):
        return self.__to

    def set_from(self, from_):
        self.__from = from_

    def get_from(self):
        return self.__from

    def set_id(self, id_):
        self.__id = id_

    def get_id(self):
        return self.__id
    
    def set_type(self, type_):
        self.__type = type_
        
    def get_type(self):
        return self.__type

    def set_query(self, query_):
        self.__query = query_
        
    def get_query(self):
        return self.__query
    
    def simple_query(self, namespace, tag, data = "", **attrs):
        e = Element("%s %s" % (namespace, tag))
        e.cdata = data

        for i in attrs.items():
            e[i[0]] = i[1]

        self.__query = e
        
    def simple_param(self, tag, data = "", **attrs):
        e = Element(tag)
        e.cdata = data

        for i in attrs.items():
            e[i[0]] = i[1]

        self.__query.add_child(e)
        
    def set_error(self, type, msg):
        if self.__type and self.__type != "error":
            raise BadUsage, "Cant convert a %s query into an error" % self.__type
        self.__error = (type, msg)
        
    def get_error(self):
        return self.__error
    
class Message:
    TAG = "jabber:client message"
    
    BODY_TAG = "jabber:client body"
    ERROR_TAG = "jabber:client error"
    SUBJECT_TAG = "jabber:client subject"
    THREAD_TAG = "jabber:client thread"
    PRIORITY_TAG = "jabber:client priority"
    BODY = re.compile(BODY_TAG)
    ERROR = re.compile(ERROR_TAG)
    SUBJECT = re.compile(SUBJECT_TAG)
    THREAD = re.compile(THREAD_TAG)
    PRIORITY = re.compile(PRIORITY_TAG)
    X = re.compile(r"\S+ x")

    extension_packets = {}
    
    def __init__(self, tree = None):
        if not tree:
            self.__to = ""
            self.__from = ""
            self.__id = ""
            self.__type = "normal"
            self.__subject = ""
            self.__thread = ""
            self.__priority = ""
            self.__body = Element(Message.BODY_TAG)
            self.__x = []
            self.__error = None
        else:
            if tree.has_key("to"): self.__to = tree["to"]
            else: self.__to = ""
            if tree.has_key("from"): self.__from = tree["from"]
            else: self.__from = ""
            if tree.has_key("id"): self.__id = tree["id"]
            else: self.__id = ""
            if tree.has_key("type"): self.__type = tree["type"]
            else: self.__type = ""

            if tree.has_child(Message.BODY):
                self.__body = tree.get_children(Message.BODY)[0]
            else: self.__body = Element(Message.BODY_TAG)

            if tree.has_child(Message.ERROR):
                err = tree.get_children(Message.ERROR)[0]
                self.__error = (err["code"], err.cdata)
            else:
                self.__error = None

            if tree.has_child(Message.SUBJECT):
                self.__subject = tree.get_children(Message.SUBJECT)[0].cdata
            else:
                self.__subject = ""

            if tree.has_child(Message.THREAD):
                self.__thread = tree.get_children(Message.THREAD)[0].cdata
            else:
                self.__thread = ""

            if tree.has_child(Message.PRIORITY):
                self.__priority = tree.get_children(Message.PRIORITY)[0].cdata
            else:
                self.__priority = ""

            self.__x = []
            xen = tree.get_children(Message.X)
            for x in xen:
                if Message.extension_packets.has_key(x.type):
                    self.__x.append(Message.extension_packets[x.type](x))
                else:
                    self.__x.append(DefaultExtension(x))

    def element_tree(self):
        tree = Element(Message.TAG)
        if self.__to: tree["to"] = self.__to
        if self.__from: tree["from"] = self.__from
        if self.__id: tree["id"] = self.__id
        if self.__type: tree["type"] = self.__type

        if self.__subject:
            e = Element(Message.SUBJECT_TAG)
            e.cdata = self.__subject
            tree.add_child(e)

        if self.__thread:
            e = Element(Message.THREAD_TAG)
            e.cdata = self.__thread
            tree.add_child(e)

        if self.__priority:
            e = Element(Message.PRIORITY_TAG)
            e.cdata = self.__priority
            tree.add_child(e)
        
        tree.add_child(self.__body)
        
        for x in self.__x:
            tree.add_child(x.element_tree())
            
        if self.__error:
            e = Element(Message.ERROR_TAG)
            e["code"] = self.__error[0]
            e.cdata = self.__error[1]
            tree.add_child(e)

        return tree

    def set_to(self, to_):
        self.__to = to_

    def get_to(self):
        return self.__to

    def set_from(self, from_):
        self.__from = from_

    def get_from(self):
        return self.__from

    def set_id(self, id_):
        self.__id = id_

    def get_id(self):
        return self.__id

    def set_type(self, type_):
        if type_ in ("normal", "error", "chat", "groupchat"):
            self.__type = type_

    def get_type(self):
        return self.__type

    def set_subject(self, subj_):
        self.__subject = subj_

    def get_subject(self):
        return self.__subject

    def set_thread(self, thread_):
        self.__thread = thread_

    def get_thread(self):
        return self.__thread

    def set_priority(self, pri_):
        self.__priority = pri_

    def get_priority(self):
        return self.__priority

    def set_body(self, body_):
        self.__body = body_

    def set_body_text(self, txt):
        self.__body.cdata = txt
        
    def add_body_element(self, elt_):
        self.__body.add_child(elt_)
    
    def get_body(self):
        return self.__body

    def get_body_text(self):
        return self.__body.cdata
    
    def get_body_flat(self, pretty = 0):
        return self.__body.output()
    
    def add_extension(self, x):
        self.__x.append(x)

    def get_extensions(self):
        return self.__x

    def set_error(self, type, message):
        self.__error = (type, message)

    def get_error(self):
        return self.__error


def register_message_extension(namespace, packet):
    Message.extension_packets["%s x" % namespace] = packet

def register_presence_extension(namespace, packet):
    Presence.extension_packets["%s x" % namespace] = packet
    
def auto_register_jabber_packets(pq):
    register_message_extension(Ident.NS, Ident)
    register_message_extension(Delay.NS, Delay)
    pq.register_packet_type(Message.TAG, Message)
    pq.register_packet_type(IQ.TAG, IQ)
    pq.register_packet_type(Presence.TAG, Presence)
    pq.register_packet_type(Stream.TAG, Stream)
